from __future__ import unicode_literals

from django.db import models
from django.db.models.signals import post_delete
from django.dispatch.dispatcher import receiver
from django.contrib.auth.models import User, Group
from .validators import alphanumeric, numeric
from django.core.exceptions import ValidationError


class Organization(models.Model):
    name = models.CharField(max_length=255, default=None)
    subscription = models.CharField(
        max_length=1, choices=(
            ('F', 'Free'), ('P', 'Pro'), ))
    owner = models.ForeignKey(User)

    def __unicode__(self):
        return u'%s' % (self.name)


class UserMeta(models.Model):
    user = models.OneToOneField(User, db_index=True)
    organization = models.ForeignKey(Organization, blank=True, null=True)
    role = models.CharField(
        max_length=1, choices=(
            ('A', 'Admin'), ('V', 'Viewer'), ))


def _user_meta(self, organization=None):
    return UserMeta.objects.get_or_create(
        user=self, defaults={
            'organization': organization})[0]


User.add_to_class('bom_profile', _user_meta)


class PartClass(models.Model):
    code = models.CharField(max_length=3, unique=True)
    name = models.CharField(max_length=255, default=None)
    comment = models.CharField(max_length=255, default=None, blank=True)

    def __unicode__(self):
        return u'%s' % (self.code + ': ' + self.name)


class Manufacturer(models.Model):
    name = models.CharField(max_length=128, default=None)
    organization = models.ForeignKey(Organization)

    class Meta:
        ordering = ['name']

    def __unicode__(self):
        return u'%s' % (self.name)


# Numbering scheme is hard coded for now, may want to change this to a
# setting depending on a part numbering scheme
class Part(models.Model):
    organization = models.ForeignKey(Organization)
    
    number_class = models.ForeignKey(
        PartClass, default=None, related_name='number_class')
    number_item = models.CharField(max_length=4, default=None, blank=True, validators=[numeric])
    number_variation = models.CharField(max_length=2, default=None, blank=True, validators=[alphanumeric])
    
    description = models.CharField(max_length=255, default=None)
    revision = models.CharField(max_length=2)
    manufacturer_part_number = models.CharField(
        max_length=128, default='', blank=True)
    manufacturer = models.ForeignKey(
        Manufacturer, default=None, blank=True, null=True)
    subparts = models.ManyToManyField(
        'self',
        blank=True,
        symmetrical=False,
        through='Subpart',
        through_fields=(
            'assembly_part',
            'assembly_subpart'))

    class Meta():
        unique_together = ['number_class', 'number_item', 'number_variation']

    def full_part_number(self):
        return "{0}-{1}-{2}".format(self.number_class.code,
                                    self.number_item, self.number_variation)

    # def distributor_parts(self):
    #     return SellerPart.objects.filter(
    #         part=self).order_by(
    #         'seller',
    #         'minimum_order_quantity')

    def seller_parts(self):
        return SellerPart.objects.filter(
            part=self).order_by(
            'seller', 'minimum_order_quantity')

    def where_used(self):
        used_in_subparts = Subpart.objects.filter(assembly_subpart=self)
        used_in_parts = [subpart.assembly_part for subpart in used_in_subparts]
        return used_in_parts

    def where_used_full(self):
        def where_used_given_part(used_in_parts, part):
            where_used = part.where_used()
            used_in_parts.update(where_used)
            for p in where_used:
                where_used_given_part(used_in_parts, p)
            return used_in_parts

        used_in_parts = set()
        where_used_given_part(used_in_parts, self)
        return list(used_in_parts)

    def files(self):
        partfiles = PartFile.objects.filter(part=self)
        return partfiles

    def indented(self):
        def indented_given_bom(bom, part, qty=1, indent_level=0, subpart=None):
            bom.append({
                'part': part,
                'quantity': qty,
                'indent_level': indent_level,
                'subpart': subpart,
            })

            indent_level = indent_level + 1
            if(len(part.subparts.all()) == 0):
                return
            else:
                for sp in part.subparts.all():
                    subparts = Subpart.objects.filter(
                        assembly_part=part, assembly_subpart=sp)
                    # since assembly_part and assembly_subpart are not unique together in a Subpart
                    # there is a possibility that there are two (or more) separate Subparts of the
                    # same Part, thus we filter and iterate again
                    for subpart in subparts:
                        qty = subpart.count
                        indented_given_bom(bom, sp, qty, indent_level, subpart)

        bom = []
        cost = 0
        indented_given_bom(bom, self)
        return bom

    def optimal_seller(self, quantity=1000):
        sellerparts = SellerPart.objects.filter(part=self)
        seller = None
        for sellerpart in sellerparts:
            if (sellerpart.minimum_order_quantity <= quantity and (
                    seller is None or
                    sellerpart.unit_cost < seller.unit_cost) and
                    sellerpart.unit_cost is not None):
                seller = sellerpart
            elif seller is None:
                seller = sellerpart

        return seller

    def save(self, **kwargs):
        if self.number_item is None or self.number_item == '':
            last_number_item = Part.objects.filter(
                number_class=self.number_class, 
                organization=self.organization).order_by('number_item').last()
            if not last_number_item:
                self.number_item = '0001'
            else:
                self.number_item = "{0:0=4d}".format(
                    int(last_number_item.number_item) + 1)
        if self.number_variation is None or self.number_variation == '':
            last_number_variation = Part.objects.all().filter(
                number_class=self.number_class,
                number_item=self.number_item).order_by('number_variation').last()
            if not last_number_variation:
                self.number_variation = '01'
            else:
                self.number_variation = "{0:0=2d}".format(
                    int(last_number_variation.number_variation) + 1)
        if self.manufacturer_part_number == '' and self.manufacturer is None and self.organization is not None:
            self.manufacturer_part_number = self.full_part_number()
            self.manufacturer, c = Manufacturer.objects.get_or_create(
                name__iexact=self.organization.name)
        super(Part, self).save()

    def __unicode__(self):
        return u'%s' % (self.full_part_number())


class Subpart(models.Model):
    assembly_part = models.ForeignKey(
        Part, related_name='assembly_part', null=True)
    assembly_subpart = models.ForeignKey(
        Part, related_name='assembly_subpart', null=True)
    count = models.IntegerField(default=1)

    def clean(self):
        unusable_parts = self.assembly_part.where_used()
        if self.assembly_subpart in unusable_parts:
            raise ValidationError(_('Recursive relationship: cannot add a subpart to a part that uses itsself.'))
        if self.assembly_subpart == self.assembly_part:
            raise ValidationError(_('Recursive relationship: cannot add a subpart to itsself.'))


class Seller(models.Model):
    organization = models.ForeignKey(Organization)
    name = models.CharField(max_length=128, default=None)

    def __unicode__(self):
        return u'%s' % (self.name)


class SellerPart(models.Model):
    seller = models.ForeignKey(Seller)
    part = models.ForeignKey(Part)
    minimum_order_quantity = models.IntegerField(null=True, blank=True)
    minimum_pack_quantity = models.IntegerField(null=True, blank=True)
    unit_cost = models.DecimalField(
        null=True,
        max_digits=8,
        decimal_places=4,
        blank=True)
    lead_time_days = models.IntegerField(null=True, blank=True)
    nre_cost = models.DecimalField(
        null=True,
        max_digits=8,
        decimal_places=4,
        blank=True)
    ncnr = models.BooleanField(default=False)

    class Meta():
        unique_together = [
            'seller',
            'part',
            'minimum_order_quantity',
            'unit_cost']


class PartFile(models.Model):
    file = models.FileField(upload_to='partfiles/')
    upload_date = models.DateField(auto_now=True)
    part = models.ForeignKey(Part)


@receiver(post_delete, sender=PartFile)
def partfile_post_delete_handler(sender, instance, **kwargs):
    instance.file.delete(False)



from bom.octopart_parts_match import match_part
from bom.models import Part, PartClass, Seller, SellerPart, Subpart, \
    Manufacturer, Organization, PartFile


def create_a_fake_organization(user, free=False):
    org = Organization(
        name="Atlas",
        subscription='F' if free else 'P',
        owner=user)
    org.save()

    return org


def create_some_fake_part_classes():
    pc1 = PartClass(code=500, name='Wendy', comment='Mechanical Switches')
    pc1.save()

    pc2 = PartClass(code=200, name='Archibald', comment='')
    pc2.save()

    pc3 = PartClass(code=503, name='Ghost', comment='Like Kasper')
    pc3.save()

    return pc1, pc2, pc3


def create_a_fake_subpart(assembly_part, assembly_subpart, count=4):
    sp1 = Subpart(
        assembly_part=assembly_part,
        assembly_subpart=assembly_subpart,
        count=count)
    sp1.save()

    return sp1


def create_some_fake_sellers(organization):
    s1 = Seller(name='Mouser', organization=organization)
    s1.save()

    s2 = Seller(name='Digi-Key', organization=organization)
    s2.save()

    s3 = Seller(name='Archibald', organization=organization)
    s3.save()

    return s1, s2, s3


def create_some_fake_manufacturers(organization):
    m1 = Manufacturer(name='STMicroelectronics', organization=organization)
    m1.save()

    m2 = Manufacturer(name='Nordic Semiconductor', organization=organization)
    m2.save()

    m3 = Manufacturer(name='Murata', organization=organization)
    m3.save()

    return m1, m2, m3


def create_a_fake_seller_part(
        seller,
        part,
        moq,
        mpq,
        unit_cost,
        lead_time_days):
    sp1 = SellerPart(
        seller=seller,
        part=part,
        minimum_order_quantity=moq,
        minimum_pack_quantity=mpq,
        unit_cost=unit_cost,
        lead_time_days=lead_time_days)
    sp1.save()

    return sp1


def create_some_fake_parts(organization):
    (pc1, pc2, pc3) = create_some_fake_part_classes()
    (m1, m2, m3) = create_some_fake_manufacturers(organization=organization)

    pt1 = Part(
        manufacturer_part_number='STM32F401CEU6',
        number_class=pc2,
        number_item='3333',
        description='Brown dog',
        revision='1',
        manufacturer=m1,
        organization=organization)
    pt1.save()

    pt2 = Part(
        manufacturer_part_number='GRM1555C1H100JA01D',
        number_class=pc1,
        description='',
        manufacturer=None,
        organization=organization)
    pt2.save()

    pt3 = Part(
        manufacturer_part_number='NRF51822',
        number_class=pc3,
        description='Friendly ghost',
        manufacturer=m3,
        organization=organization)
    pt3.save()

    create_a_fake_subpart(pt1, pt2)
    create_a_fake_subpart(pt1, pt3, count=10)

    (s1, s2, s3) = create_some_fake_sellers(organization=organization)

    create_a_fake_seller_part(
        s1,
        pt1,
        moq=None,
        mpq=None,
        unit_cost=None,
        lead_time_days=None)
    create_a_fake_seller_part(
        s2,
        pt1,
        moq=1000,
        mpq=5000,
        unit_cost=0.1005,
        lead_time_days=7)
    create_a_fake_seller_part(
        s2,
        pt2,
        moq=200,
        mpq=200,
        unit_cost=0,
        lead_time_days=47)

    return pt1, pt2, pt3


def create_a_fake_partfile(file, part):
    pf1 = PartFile(file=None, part=part)
    pf1.save()

    return pf1



from django.conf.urls import url
from django.contrib.auth import views as auth_views
from django.contrib import admin
from django.views.generic import TemplateView

from . import views

urlpatterns = [
    # these will likely be overridden by your app
    url(r'^admin/', admin.site.urls),
    url(r'^login/$', auth_views.login, {'redirect_authenticated_user': True}, name='login'),
    url(r'^logout/$', auth_views.logout, {'next_page': '/'}, name='logout'),
    url(r'^about/$', TemplateView.as_view(template_name='error.html'), name='about'), # TODO: remove this and make tests pass..

    url(r'^$', views.home, name='home'),
    url(r'^error/$', views.error, name='error'),
    url(r'^signup/$', views.bom_signup, name='bom-signup'),
    url(r'^export/$', views.export_part_list, name='export-part-list'),
    url(r'^create-part/$', views.create_part, name='create-part'),
    url(r'^upload-parts/$', views.upload_parts, name='upload-parts'),
    url(r'^part/(?P<part_id>[0-9]+)/$', views.part_info, name='part-info'),
    url(r'^part/(?P<part_id>[0-9]+)/export/$', views.part_export_bom, name='part-export-bom'),
    url(r'^part/(?P<part_id>[0-9]+)/upload/$', views.part_upload_bom, name='part-upload-bom'),
    url(r'^part/(?P<part_id>[0-9]+)/octopart-match/$', views.part_octopart_match, name='part-octopart-match'),
    url(r'^part/(?P<part_id>[0-9]+)/octopart-match-indented/$', views.part_octopart_match_bom, name='part-octopart-match-bom'),
    url(r'^part/(?P<part_id>[0-9]+)/edit/$', views.part_edit, name='part-edit'),
    url(r'^part/(?P<part_id>[0-9]+)/delete/$', views.part_delete, name='part-delete'),
    url(r'^part/(?P<part_id>[0-9]+)/manage-bom/$', views.manage_bom, name='part-manage-bom'),
    url(r'^part/(?P<part_id>[0-9]+)/add-subpart/$', views.add_subpart, name='part-add-subpart'),
    url(r'^part/(?P<part_id>[0-9]+)/add-sellerpart/$', views.add_sellerpart, name='part-add-sellerpart'),
    url(r'^part/(?P<part_id>[0-9]+)/upload-file/$', views.upload_file_to_part, name='part-upload-partfile'),
    url(r'^part/(?P<part_id>[0-9]+)/delete-file/(?P<partfile_id>[0-9]+)/$', views.delete_file_from_part, name='part-delete-partfile'),
    url(r'^part/(?P<part_id>[0-9]+)/remove-all-subparts/$', views.remove_all_subparts, name='part-remove-all-subparts'),
    url(r'^part/(?P<part_id>[0-9]+)/remove-subpart/(?P<subpart_id>[0-9]+)/$', views.remove_subpart, name='part-remove-subpart'),
]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-06-16 04:20
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('bom', '0001_squashed_0018_auto_20170418_0330'),
    ]

    operations = [
        migrations.AlterField(
            model_name='usermeta',
            name='organization',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='bom.Organization'),
        ),
    ]
